// A map data structure that holds object keys weakly, yet can also hold
// non-object keys, unlike the native WeakMap.
var UniversalWeakMap = function UniversalWeakMap() {
  // Since a WeakMap cannot hold primitive values as keys, we need a
  // backup Map instance to hold primitive keys. Both this._weakMap and
  // this._strongMap are lazily initialized.
  this._weakMap = null;
  this._strongMap = null;
};

// Since get and set are the only methods used, that's all I've
// implemented here.

UniversalWeakMap.prototype.get = function get (key) {
  var map = this._getMap(key, false);
  if (map) {
    return map.get(key);
  }
};

UniversalWeakMap.prototype.set = function set (key, value) {
  this._getMap(key, true).set(key, value);
  // An actual Map or WeakMap would return this here, but returning the
  // value is more convenient below.
  return value;
};

UniversalWeakMap.prototype._getMap = function _getMap (key, canCreate) {
  if (! canCreate) {
    return isObjRef(key) ? this._weakMap : this._strongMap;
  }
  if (isObjRef(key)) {
    return this._weakMap || (this._weakMap = new WeakMap);
  }
  return this._strongMap || (this._strongMap = new Map);
};

function isObjRef(value) {
  switch (typeof value) {
  case "object":
    if (value === null) {
      return false;
    }
  case "function":
    return true;
  default:
    return false;
  }
}

var root = new UniversalWeakMap;

function tuple() {
  var items = [], len = arguments.length;
  while ( len-- ) items[ len ] = arguments[ len ];

  return Tuple.intern(items);
}

var Tuple = function Tuple() {
  var items = [], len = arguments.length;
  while ( len-- ) items[ len ] = arguments[ len ];

  return Tuple.intern(items, this);
};

Tuple.intern = function intern (array, tuple) {
  var node = root;

  array.forEach(function (item) {
    node = node.get(item) || node.set(item, new UniversalWeakMap);
  });

  if (node.tuple) {
    return node.tuple;
  }

  tuple = tuple || Object.create(Tuple.prototype);

  array.forEach(function (item, i) { return def(tuple, i, item, true); });
  def(tuple, "length", array.length, false);

  return node.tuple = tuple;
};

Tuple.from = function from (iterable) {
  return Tuple.intern([].concat( iterable ));
};

function def(obj, name, value, enumerable) {
  Object.defineProperty(obj, name, {
    value: value,
    enumerable: !! enumerable,
    writable: false,
    configurable: false
  });
}

export { tuple, Tuple };
